import matplotlib.pyplot as plt
import numpy as np

import torch
import torchvision
import torchvision.transforms as transforms

import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

# this is a command line program which can be run with different options
import argparse

from zmq import device
parser = argparse.ArgumentParser()
parser.add_argument("-b","--batch_size", type=int, default=32, help="batch size")
parser.add_argument("--optimizer", type=str, default="SGD", help="optimizer")
parser.add_argument("--lr", type=float, default=0.001, help="learning rate")
parser.add_argument("-m", "--momentum", type=float, default=0.9, help="momentum")
parser.add_argument("--epochs", type=int, default=1, help="training epochs")


class Data():
    def __init__(self, args):
        # transforms
        transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize((0.5,), (0.5,))])        
        self.train_set = torchvision.datasets.MNIST('../data', download=True, train=True, transform=transform)
        self.test_set = torchvision.datasets.MNIST('../data', download=True, train=False, transform=transform)
        # dataloaders
        self.train_loader = torch.utils.data.DataLoader(self.train_set, batch_size=args.batch_size, shuffle=True, num_workers=0)
        self.test_loader = torch.utils.data.DataLoader(self.test_set, batch_size=args.batch_size, shuffle=True, num_workers=0)
        # Task: split train_set into train_set and val_set and create loaders

def new_network(args):
    net = nn.Sequential(nn.Flatten(), nn.Linear(28 * 28, 10))
    net.to(args.dev)
    return net

def save_network(net:nn.Module, filename):
    torch.save({'model_state_dict': net.state_dict()}, filename)

def load_network(args, filename):
    state = torch.load(open(filename, "rb") , map_location=args.dev)
    net = new_network(args)
    net.load_state_dict(state['model_state_dict'])
    return net


def evaluate(net: nn.Module, loader):
    ## Task: perform evaluation of the network on the data given by the loader
    net.eval()
    loss = 0
    acc = 0
    return loss, acc


def train(args):
    print(args)
    data = Data(args)

    # lets verify how the loader packs the data
    (input, target) = next(iter(data.train_loader))
    # expect to get [batch_size x 1 x 28 x 28]
    print('Input  type:', input.type())
    print('Input  size:', input.size())
    # xpect to get [batch_size]
    print('Labels size:', target.size())
    # see number of trainig data:
    n_train_data = len(data.train_set)
    print('Train data size:', n_train_data)

    # network, expect input images 28* 28 and 10 classes
    net = new_network(args)

    # loss function
    loss = nn.CrossEntropyLoss(reduction='none')

    # optimizer
    optimizer = optim.SGD(net.parameters(), lr=args.lr, momentum=args.momentum)

    for epoch in range(args.epochs):
        # will accumulate total loss over the dataset
        L = 0
        # loop fetching a mini-batch of data at each iteration
        for i, (x, target) in enumerate(data.train_loader):
            x = x.to(dev)
            target = target.to(dev)
            # apply the network
            y = net.forward(x)
            # calculate mini-batch losses
            l = loss(y, target)
            # accumulate the total loss as a regular float number (important to sop graph tracking)
            L += l.sum().item()
            # the gradient usually accumulates, need to clear explicitly
            optimizer.zero_grad()
            # compute the gradient from the mini-batch loss
            l.mean().backward()
            # make the optimization step
            optimizer.step()
        print(f'Epoch: {epoch} mean loss: {L / n_train_data}')
        save_network(net, 'network.pt')


if __name__ == "__main__":
    args = parser.parse_args()
    # query if we have GPU
    dev = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
    print('Using device:', dev)
    args.dev = dev
    train(args)